#  ***** GPL LICENSE BLOCK *****
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  All rights reserved.
#  ***** GPL LICENSE BLOCK *****

import bpy, bmesh
import struct
from collections import namedtuple


# type definitions
BSPHeader = namedtuple('BSPHeader', 
    ("version,"
    "entities_ofs,  entities_size,"
    "planes_ofs,    planes_size,"
    "miptex_ofs,    miptex_size,"
    "verts_ofs,     verts_size,"
    "visilist_ofs,  visilist_size,"
    "nodes_ofs,     nodes_size,"
    "texinfo_ofs,   texinfo_size,"
    "faces_ofs,     faces_size,"
    "lightmaps_ofs, lightmaps_size,"
    "clipnodes_ofs, clipnodes_size,"
    "leaves_ofs,    leaves_size,"
    "lface_ofs,     lface_size,"
    "edges_ofs,     edges_size,"
    "ledges_ofs,    ledges_size,"
    "models_ofs,    models_size")
    )
fmt_BSPHeader = '<31i'

BSPModel = namedtuple('BSPModel',
    ("bbox_min_x,  bbox_min_y,  bbox_min_z,"
    "bbox_max_x,   bbox_max_y,  bbox_max_z,"
    "origin_x,     origin_y,    origin_z,"
    "node_id0, node_id1, node_id2, node_id3,"
    "numleafs,"
    "face_id,"
    "face_num")
    )
fmt_BSPModel = '<9f7i'

BSPFace = namedtuple('BSPFace',
    ("plane_id,"
    "size,"
    "ledge_id,"
    "ledge_num,"
    "texinfo_id,"
    "lighttype,"
    "lightlevel,"
    "light0, light1,"
    "lightmap")
    )
fmt_BSPFace = '<HHiHHBBBBi'

BSPVertex = namedtuple('BSPVertex', 'x, y, z')
fmt_BSPVertex = '<fff'

BSPEdge = namedtuple('BSPEdge', 'vertex0, vertex1')
fmt_BSPEdge = '<HH'

BSPTexInfo = namedtuple('BSPTexInfo',
    ("s_x,  s_y,    s_z,    s_ofs,"
    "t_x,   t_y,    t_z,    t_ofs,"
    "texture_id,"
    "animated")
    )
fmt_BSPTexInfo = '<8f2I'


# functions
def mesh_add(mesh_id):
    if bpy.context.scene.objects.active:
        bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.add(type='MESH', enter_editmode=True)
    ob = bpy.context.object
    ob.name = "bsp_model_%d" % mesh_id
    ob.show_name = True
    ob_mesh = ob.data
    ob_mesh.name = ob.name + "_mesh"
    return ob

def mesh_append_face(bm, vertex_list):
    scale = 0.05
    for vert in reversed(vertex_list):
        bm.verts.new((vert[0] * scale, vert[1] * scale, vert[2] * scale))
    bm.faces.new((bm.verts[i] for i in range(-len(vertex_list),0)))

def import_bsp(context, filepath):
    with open(filepath, 'rb') as file:
        header_data = file.read(struct.calcsize(fmt_BSPHeader))
        header = BSPHeader._make(struct.unpack(fmt_BSPHeader, header_data))

        num_models = int(header.models_size / struct.calcsize(fmt_BSPModel))
        num_faces = int(header.faces_size / struct.calcsize(fmt_BSPFace))
        num_edges = int(header.edges_size / struct.calcsize(fmt_BSPEdge))
        num_verts = int(header.verts_size / struct.calcsize(fmt_BSPVertex))

        print("Importing from \'%s\'" % filepath)
        print("Map contains %d models. (faces = %d, edges = %d, verts = %d)" % (num_models, num_faces, num_edges, num_verts))

        # read models, faces, edges and vertices into buffers
        file.seek(header.models_ofs)
        model_data = file.read(header.models_size)
        file.seek(header.faces_ofs)
        face_data = file.read(header.faces_size)
        file.seek(header.edges_ofs) # actual edges
        edge_data = file.read(header.edges_size)

        # read in the list of edges and store in readable form (flat list of ints)
        file.seek(header.ledges_ofs) 
        edge_index_list = struct.unpack('<'+int(header.ledges_size/4)*'i', file.read(header.ledges_size))
        # do the same with vertices (flat list of floats)
        file.seek(header.verts_ofs)
        vertex_list = struct.unpack('<'+int(header.verts_size/4)*'f', file.read(header.verts_size))

    # create some structs for storing data
    model_size = struct.calcsize(fmt_BSPModel)
    model_struct = struct.Struct(fmt_BSPModel)
    face_size = struct.calcsize(fmt_BSPFace)
    face_struct = struct.Struct(fmt_BSPFace)
    edge_size = struct.calcsize(fmt_BSPEdge)
    edge_struct = struct.Struct(fmt_BSPEdge)
    vertex_size = struct.calcsize(fmt_BSPVertex)
    vertex_struct = struct.Struct(fmt_BSPVertex)

    for m in range(0, num_models):
        print("---- MODEL %d:" % m, end = ' ')
        model_ofs = m * model_size
        model = BSPModel._make(model_struct.unpack_from(model_data[model_ofs:model_ofs+model_size]))
        print("(%d)" % model.face_num)

        # create new mesh
        mesh_object = mesh_add(m)
        bm = bmesh.from_edit_mesh(mesh_object.data)

        for f in range(0, model.face_num):
            # print("FACE %d:" % f, end = ' ')
            face_ofs = (model.face_id + f) * face_size
            face = BSPFace._make(face_struct.unpack_from(face_data[face_ofs:face_ofs+face_size]))
            # print("(%d)" % face.ledge_num)

            # populate a list with vertices
            face_vertices = []

            for i in range(0,face.ledge_num):
                edge_index = edge_index_list[face.ledge_id+i]
                if edge_index < 0:
                    # vertex order is 1->0
                    edge_ofs = -edge_index * edge_size
                    edge = BSPEdge._make(edge_struct.unpack_from(edge_data[edge_ofs:edge_ofs+edge_size]))
                    vofs = edge[1] * 3
                    face_vertices.append( [vertex_list[vofs], vertex_list[vofs+1], vertex_list[vofs+2]] )
                else:
                    # vertex order is 0->1
                    edge_ofs = edge_index * edge_size
                    edge = BSPEdge._make(edge_struct.unpack_from(edge_data[edge_ofs:edge_ofs+edge_size]))
                    vofs = edge[0] * 3
                    face_vertices.append( [vertex_list[vofs], vertex_list[vofs+1], vertex_list[vofs+2]] )
            # print(face_vertices)

            # append face to mesh
            mesh_append_face(bm, face_vertices)

        # update the mesh with data from the bmesh
        bmesh.update_edit_mesh(mesh_object.data)